#! /usr/bin/env node

// const inquirer = require('inquirer');
import inquirer from "inquirer";
import chalk from "chalk";
import { program } from "commander";
import pkg from "../package.json";
import fs from "fs";
import path from "path";
import ejs from "ejs";
import {
  install_deps,
  run_build,
  run_git_init,
  run_next,
  run_watch,
} from "./runtime";
import { writeGitIgnore } from "./files";
import { publish_dapp } from "./publish";
import { unpublish } from "./unpublish";
import { start_server } from "./server";

export const a = 0;

/**
 * 渲染模板
 * @param {string} template_dir 模板文件夹
 * @param {string} dest_dir 模板文件夹
 * @param {object} data
 */
async function render(template_dir: string, dest_dir: string, data: object) {
  const files = fs.readdirSync(template_dir);
  for (const file of files) {
    // 使用 ejs 渲染对应的模版文件
    // renderFile（模版文件地址，传入渲染数据）
    const tmplPath = path.join(template_dir, file);
    const destPath = path.join(dest_dir, file);
    if (fs.statSync(tmplPath).isDirectory()) {
      fs.mkdirSync(destPath);
      await render(tmplPath, destPath, data);
    } else {
      const extension = tmplPath.split('.').pop()
      if (['png', 'jpg', 'ico', 'svg'].includes(extension!)) {
        fs.writeFileSync(destPath, fs.readFileSync(tmplPath));
        continue
      }
      ejs.renderFile(tmplPath, data).then((content) => {
        // 生成 ejs 处理后的模版文件
        fs.writeFileSync(path.join(dest_dir, file), content);
      });
    }
  }
}

const __dirname = path.dirname(__filename);

// 名称，描述，版本号，用法提示。
program
  .name("cpc-cli")
  .description(
    "This is a CLI command, to help developers develop CPChain dapps"
  )
  .version(`cpc-cli ${pkg.version}`)
  .usage("<command> [options]");

// Create dapps command
program
  .command("create <name>")
  .description("Create DApp") // 命令描述
  .action((name, options) => {
    const arr = [
      {
        type: "input",
        name: "projectName",
        message: "Project Name",
        default: name,
      },
      {
        type: "list",
        name: "projectType",
        message: "Project Type",
        default: "Web DApp",
        choices: [
          { name: "Web DApp", value: "web-dapp" },
          { name: "Docs", value: "docs" },
        ],
      },
      {
        type: "confirm",
        name: "confirm",
        message: "Confirm",
      },
    ];
    inquirer
      .prompt(arr)
      .then(async (answers) => {
        if (!answers.confirm) {
          return;
        }
        // 生成文件目录
        const cwdUrl = process.cwd();
        const projectDir = path.join(
          cwdUrl,
          answers.projectName.replace(" ", "-")
        );
        // 检查文件夹是否存在
        if (!fs.existsSync(projectDir)) {
          fs.mkdirSync(projectDir);
        }
        // 检查文件夹是否为空
        if (fs.readdirSync(projectDir).length !== 0) {
          throw new Error(`Aborted, becuase ${projectDir} is not empty`);
        }

        // templates
        const templates: { [key: string]: string } = {
          "web-dapp": "../templates/app-with-tw",
          docs: "../templates/default-docs",
        };

        // 模版文件目录
        const destUrl = path.join(__dirname, templates[answers.projectType]);
        // 从模版目录中读取文件
        await render(destUrl, projectDir, answers);
        // git init
        await run_git_init(projectDir);
        // write gitignore file
        writeGitIgnore(projectDir);
        // Notes
        console.log(
          "----------------------------------------------------------------\nGet started:\n"
        );
        console.log(chalk.green("cd"), answers.projectName);
        console.log(chalk.green("yarn"), "build");
        console.log(chalk.green("cpc-cli"), "run");
        console.log(
          "----------------------------------------------------------------"
        );
      })
      .catch((error) => {
        console.error(error);
      });
  });

// Run
program
  .command("run")
  .option('--taobao', "use taobao registry", false)
  .description("Start to run this project")
  .action(async (options: {taobao: boolean}) => {
    console.log(chalk.green("Start..."));
    // 当前目录
    const cwdUrl = process.cwd();
    // 创建 .cpchain
    const cpchainDir = ".cpchain";
    if (!fs.existsSync(cpchainDir)) {
      fs.mkdirSync(cpchainDir);
    }

    if (!fs.existsSync(path.join(cpchainDir, "package.json"))) {
      // 初始化 runtime
      // 复制 Runtime
      const templates: { [key: string]: string } = {
        "web-dapp": "../runtimes/dapp",
        docs: "../runtimes/docs",
      };
      const destUrl = path.join(__dirname, templates["web-dapp"]);
      await render(destUrl, cpchainDir, {});
    }
    const taobao_registry = 'https://registry.npm.taobao.org'
    await install_deps(cwdUrl, options.taobao ? taobao_registry: undefined);
    await install_deps(cpchainDir, options.taobao ? taobao_registry: undefined);
    await run_build(cwdUrl); // build dapp
    await run_watch(cwdUrl);
  });

program
  .command("publish")
  .description("Publish this web-dapps")
  .option("-c, --cpm-url <url>", "CPM URL", "https://cpm.cpchain.io")
  .option("-p, --http-proxy <url>", "HTTP Proxy URL")
  .action(async (options: { cpmUrl: string; httpProxy: string }) => {
    // 检查当前文件夹是否有 package.json，若没有，则报错
    const pkgJson = "package.json";
    if (!fs.existsSync(pkgJson)) {
      console.log(chalk.red(`${pkgJson} does not exist`));
      return;
    }
    // 检查当前文件夹是否有 .cpchain/.dapp，若没有，则报错
    const cpchainDir = ".cpchain";
    const dappDir = path.join(cpchainDir, ".dapp");
    if (!fs.existsSync(dappDir)) {
      console.log(
        chalk.red(`Error: Please run`),
        chalk.green("`cpc-cli run`"),
        chalk.red("to build first")
      );
      return;
    }
    // read from package.json
    const pfgContent = fs
      .readFileSync(pkgJson, { encoding: "utf-8" })
      .toString();
    const pkg = JSON.parse(pfgContent);
    if (!pkg.name) {
      console.log(
        chalk.red("Error: Please set name first on your package.json")
      );
      return;
    }
    if (!pkg.version) {
      console.log(
        chalk.red("Error: Please set the version first on your package.json")
      );
      return;
    }
    console.log(chalk.green('You will publish this dapp'))
    console.log(chalk.gray('--------------------------------'))
    console.log(chalk.green('Name:'), pkg.name)
    console.log(chalk.green('Description:'), pkg.description || pkg.name)
    console.log(chalk.green('version:'), pkg.version)
    console.log(chalk.gray('--------------------------------'))
    const arr = [
      {
        type: "confirm",
        name: "confirm",
        message: "Confirm",
      },
    ];
    inquirer.prompt(arr).then(async (answers) => {
      if (!answers.confirm) {
        return;
      }
      await publish_dapp({
        name: pkg.name,
        description: pkg.description || pkg.name,
        version: pkg.version,
        distPath: dappDir,
        cpchainDir: cpchainDir,
        cpmHost: options.cpmUrl,
        httpProxy: options.httpProxy,
        logoPath: pkg.cpchain?.logo,
        faviconPath: pkg.cpchain?.favicon
      });
    });
  });

program
  .command("unpublish")
  .description("Publish this web-dapps")
  .option("-c, --cpm-url <url>", "CPM URL", "https://cpm.cpchain.io")
  .option("-p, --http-proxy <url>", "HTTP Proxy URL")
  .option("--publish", "If set publish, your dapp will be published again")
  .action(async (options: { cpmUrl: string; httpProxy: string, publish: boolean }) => {
    const pkgJson = "package.json";
    if (!fs.existsSync(pkgJson)) {
      console.log(chalk.red(`${pkgJson} does not exist`));
      return;
    }
    const pfgContent = fs
      .readFileSync(pkgJson, { encoding: "utf-8" })
      .toString();
    const pkg = JSON.parse(pfgContent);
    await unpublish({
      name: pkg.name,
      cpmHost: options.cpmUrl,
      httpProxy: options.httpProxy,
      publish: options.publish
    })
  })

program.command('server')
  .description('Start dapps server')
  .option("-c, --cdl-dir <dir>", "CDL Dir")
  .option("-f, --force-pull", "Force pull images", false)
  .option("-p, --port <port>", "Port of server to listen on", "8080")
  .action(async ({ cdlDir, port, forcePull }: { cdlDir?: string, port: string, forcePull: boolean }) => {
    if (!cdlDir) {
      console.log(chalk.red('Please provide a cdl file by passing `-c` parameter'))
      process.exit(0)
    }
    await start_server({
      host: '127.0.0.1',
      port: parseInt(port),
      cpchainDir: '.cpchain',
      targetCdlPath: cdlDir,
      forcePull
    })
  })

program.parse(process.argv);
