import { DAppInfo } from "./types";
import tar from "tar";
import path from "path";
import fs, { Dirent } from "fs";
import { request } from "./utils";
import { lookup } from "mime-types"
import chalk from "chalk";

// function to encode file data to base64 encoded string
function base64_encode(file: string) {
  // read binary data
  var bitmap = fs.readFileSync(file);
  // convert binary data to base64 encoded string
  return new Buffer(bitmap).toString("base64");
}

function read_image_file(file: string) {
  const content = base64_encode(file);
  const postfix = file.split(".").pop();
  return { content, postfix };
}

async function walkSync(currentDirPath: string, callback?: (file: string, drient: Dirent) => Promise<void>): Promise<string[]> {
  // http://nodejs.cn/api/fs.html#fsreaddirsyncpath-options
  // http://nodejs.cn/api/fs.html#class-fsdirent 新增于: v10.10.0
  const files: string[] = []
  fs.readdirSync(currentDirPath, { withFileTypes: true }).forEach(async function(dirent) {
    const filePath = path.join(currentDirPath, dirent.name);
    files.push(filePath)
    if (dirent.isFile()) {
      if (callback) {
        await callback(filePath, dirent);
      }
    } else if (dirent.isDirectory()) {
      await walkSync(filePath, callback);
    }
    return filePath
  });
  return files
}

// 遍历所有静态文件并上传，完成后，返回一个 <文件名, URL> 的映射表，供外层替换
async function publish_static_files({ editorServer = "https://editor.cpchain.io", distPath, httpProxy }: { editorServer?: string, distPath: string, httpProxy?: string}): Promise<{ [key: string]: string }> {
  // 遍历 distPath
  const allowPostfix = ['png', 'jpg', 'jpeg', 'ico', 'svg', 'gif', 'woff', 'woff2', 'pdf', 'eot', 'ttf']
  const mapping: { [key: string]: string} = {}
  const files = await walkSync(distPath)
  for (const file of files) {
    const postfix = file.split('.').pop()
    if (postfix && allowPostfix.includes(postfix)) {
      // upload file
      const fd = new FormData();
      const fileBlob = new Blob([fs.readFileSync(file)], { type: lookup(file) || undefined });
      const filename = file.split('/').pop()
      console.log(chalk.green('Uploading file:'), chalk.gray(filename))
      fd.set("file", fileBlob, filename);
      const res = await request(`${editorServer}/upload`, { httpProxy, method: "POST", body: fd })
      const data: any = await res.json()
      const url = editorServer + data.url
      mapping[file] = url
      console.log(chalk.green('Upload successfully: ' + url))
    }
  }
  return mapping
}

export async function publish_dapp({
  name,
  description,
  cpmHost,
  version,
  distPath,
  cpchainDir,
  httpProxy,
  logoPath,
  faviconPath,
}: DAppInfo) {
  // 发布静态文件
  const mapping = await publish_static_files({ distPath, httpProxy })
  // 首先 copy 一份 main.js
  const mainJsPath = path.join(distPath, 'main.js')
  const originMainJsPath = path.join(distPath, 'main.origin.js')
  fs.cpSync(mainJsPath, originMainJsPath)
  let main_js = fs.readFileSync(mainJsPath).toString()
  // 替换
  for (const f in mapping) {
    main_js = main_js.replace(f.split('/').pop() || '__$_wrong', mapping[f])
  }
  // 写回文件
  fs.writeFileSync(mainJsPath, main_js)
  console.log(chalk.gray('Publishing to DApp market...'));
  // 将 dist 中的文件打包为 tar.gz
  const main = path.join("main.js");
  const dappYml = path.join("dapp.yml");
  const packageFile = path.join(cpchainDir, `${name}-${version}.tar.gz`);
  // package
  await tar.c(
    {
      gzip: true,
      C: distPath,
      file: packageFile,
    },
    [main, dappYml]
  );
  // base64 encoded
  const encoded = base64_encode(packageFile);
  const payload: any = {
    name,
    description,
    version,
    kind: "web-dapps",
    pkg: encoded,
    author: null,
    logo: null,
  };
  if (logoPath) {
    payload.logo = read_image_file(logoPath);
  }
  if (faviconPath) {
    payload.favicon = read_image_file(faviconPath);
  }

  // post
  const response = await request(`${cpmHost}/push_version`, {
    httpProxy,
    method: "post",
    body: JSON.stringify(payload),
    headers: { "Content-Type": "application/json" },
  });
  const data: any = await response.json();
  if (!data.error) {
    console.log(chalk.green('Publish successfully\n'))
    const url = `https://web-dapps.cpchain.io/dapp/${data.id}`
    console.log(chalk.gray(`DApp Market: https://web-dapps.cpchain.io`))
    console.log(chalk.gray(`View your dapp on ${url}`))
  } else {
    console.log(chalk.gray('Error code:'), chalk.red(data.code))
    console.log(chalk.gray('Error message:'), chalk.red(data.error))
  }
  // 修改回来
  fs.cpSync(originMainJsPath, mainJsPath)
  fs.rmSync(originMainJsPath)
}
