// 启动 dapps-server 以运行 CDL
// https://docs.docker.com/engine/api/v1.30/#tag/Container/operation/ContainerCreate

import chalk from "chalk";
import { Docker } from "node-docker-api";
import { Container } from "node-docker-api/lib/container";
import { exit } from "process";
import path from "path";
import fs from "fs";
import { delay } from "./utils";
import Progrss from 'cli-progress'

const MYSQL_NAME = "cpc-cli-start-server-mysql";
const MYSQL_IMAGE = "mysql:5.7";
const MYSQL_DATABASE = "cpchain";
const MYSQL_ROOT_PASSWORD = "cpchain";
const SYNC_NAME = "cpc-cli-sync-dapps";
const SERVER_NAME = "cpc-cli-dapps-server";
const EXPLORER_IMAGE = "liaojl/explorer-v2";

async function pull_images(docker: Docker, force: boolean = false) {
  const m = await docker.image.list();
  const images: string[] = m
    .map((i) => (i.data as any).RepoTags as string[])
    .filter((i) => !!i)
    .reduce((a, b) => [...a, ...b], []);
  if (!images.includes("mysql:5.7")) {
    console.log(chalk.green("Pulling mysql:5.7 image..."));
    await docker.image.create({}, { fromImage: "mysql", tag: "5.7" });
  }
  if (!images.includes("liaojl/explorer-v2:latest") || force) {
    console.log(chalk.green("Pulling liaojl/explorer-v2:latest image..."));
    await docker.image.create(
      {},
      { fromImage: "liaojl/explorer-v2", tag: "latest" }
    );
  }
}

async function exists_container(
  docker: Docker,
  container: string
): Promise<Container | null> {
  const containers = await docker.container.list({
    all: true,
  });
  const r = containers.filter((c) => {
    const data = c.data as any;
    return data.Names.includes(`/${container}`);
  });
  if (r.length > 0) {
    return r[0];
  }
  return null;
}

async function start_mysql(docker: Docker) {
  const mysql = await exists_container(docker, MYSQL_NAME);
  if (mysql) {
    await remove(docker, "database", MYSQL_NAME);
  }
  console.log(chalk.green("Create MySQL container..."));
  const container = await docker.container.create({
    Image: MYSQL_IMAGE,
    name: MYSQL_NAME,
    Env: [
      `MYSQL_ROOT_PASSWORD=${MYSQL_ROOT_PASSWORD}`,
      `MYSQL_DATABASE=${MYSQL_DATABASE}`,
    ],
  });
  await container.start();
}

interface Props {
  host: string;
  port: number;
  cpchainDir: string;
  targetCdlPath: string;
  forcePull: boolean;
}

async function start_sync_dapps(docker: Docker, props: Props) {
  const c = await exists_container(docker, SYNC_NAME);
  if (c) {
    await remove(docker, "syncer", SYNC_NAME);
  }
  // 在 .cpchain 中创建 .cdl-dapps，将 target 复制进去
  const dir = path.join(props.cpchainDir, ".cdl-dapps");
  if (!fs.existsSync(dir)) {
    fs.mkdirSync(dir);
  }
  const filename = path.basename(props.targetCdlPath);
  const target = path.join(dir, filename);
  if (fs.statSync(props.targetCdlPath).isDirectory()) {
    fs.cpSync(props.targetCdlPath, target, { recursive: true });
  } else {
    fs.copyFileSync(props.targetCdlPath, target);
  }
  console.log(chalk.green("Create Syncer container..."));
  const container = await docker.container.create({
    Image: EXPLORER_IMAGE,
    name: SYNC_NAME,
    HostConfig: {
      Links: [`${MYSQL_NAME}:db`],
      Binds: [`${path.resolve(target)}:/explorer/dapps_dir`],
    },
    Cmd: [
      "sync-dapps-simple",
      "--mysql-url",
      "mysql://root:cpchain@db:3306/cpchain",
      "--dapps-dir",
      "/explorer/dapps_dir",
    ],
  });
  await container.start();
  // 循环检查 container status，如果停止了，说明编译失败，解析日志，抛出异常
  let cnt = 20
  const bar = new Progrss.SingleBar({
    format: chalk.gray('Syncer Progress |') + chalk.greenBright('{bar}') + chalk.gray('| {percentage}% || {value}/{total} Chunks'),
    barCompleteChar: '\u2588',
    barIncompleteChar: '\u2591',
    hideCursor: true
  }, Progrss.Presets.shades_classic);
  bar.start(100, 0)
  while (cnt--) {
    await delay(1000)
    bar.update((20 - cnt) * 5)
    const c = await exists_container(docker, SYNC_NAME)
    if (!c) {
      exit(0)
    }
    const status = await get_status(c)
    if (status !== 'running') {
      // 说明编译失败
      // 获取容器日志，输出
      const stream = await c.logs({
        follow: false,
        stdout: true,
        stderr: true
      }) as any
      stream.on('data', (info: Buffer) => {
        const logs = info.toString()
        for (const log of logs.split('\n')) {
          if (log.includes('panicked')) {
            console.log(chalk.red(log.substring(1)))
          } else {
            console.log(chalk.gray(log.substring(1)))
          }
        }
      })
      stream.on('error', (err: Buffer) => console.log(err.toString()));
      await delay(1000)
      await stop_all(docker)
      exit(0)
    }
  }
  bar.stop()
}

async function start_dapps_server(docker: Docker, port: number) {
  const c = await exists_container(docker, SERVER_NAME);
  if (c) {
    await remove(docker, "server", SERVER_NAME);
  }
  console.log(chalk.green("Create dapps-server container..."));
  const container = await docker.container.create({
    Image: EXPLORER_IMAGE,
    name: SERVER_NAME,
    ExposedPorts: {
      "8080/tcp": {}, // port inside of docker container
    },
    HostConfig: {
      Links: [`${MYSQL_NAME}:db`],
      PortBindings: {
        "8080/tcp": [
          {
            HostPort: "" + port,
          },
        ],
      },
    },
    Cmd: [
      "dapps-server",
      "--mysql-url",
      "mysql://root:cpchain@db:3306/cpchain",
      "--host",
      "0.0.0.0",
    ],
  });
  await container.start();
}

async function get_status(c: Container) {
  const status = await c.status()
  const data = status.data as any
  const state = data.State
  return state.Status
}

async function remove(docker: Docker, label: string, container: string) {
  const c = await exists_container(docker, container);
  if (c) {
    // 检查状态
    const status = await get_status(c)
    console.log(chalk.gray(`Stoppping ${label}...`));
    if (status === 'running') {
      await c.kill();
    }
    console.log(chalk.gray(`Remove ${label}...`));
    await c.delete();
  }
}

let is_stopping = false;

async function stop_all(docker: Docker) {
  if (is_stopping) {
    return;
  }
  is_stopping = true;
  await remove(docker, "syncer", SYNC_NAME);
  await remove(docker, "server", SERVER_NAME);
  await remove(docker, "database", MYSQL_NAME);
  console.log(chalk.green("Stopped all"));
  exit(0);
}

export async function start_server(props: Props) {
  let socketPath = "/var/run/docker.sock"
  // Check if platform is windows
  if (process.platform === 'win32') {
      socketPath = "//./pipe/docker_engine"
  }

  const docker = new Docker({ socketPath: socketPath });

  process.on("SIGINT", () => {
    if (is_stopping) {
      return;
    }
    console.log();
    stop_all(docker);
  }); // CTRL+C
  process.on("SIGTERM", () => {}); // `kill` command
  try {
    await pull_images(docker, props.forcePull);
    await start_mysql(docker);
    await start_sync_dapps(docker, props);
    console.log(
      chalk.gray("Syncer is syncing your contracts, please wait seconds...")
    );
    // 等待
    await delay(5000);
    await start_dapps_server(docker, props.port);
    console.log(chalk.green("Succeeded to start"));
    console.log(chalk.gray("--------------------------------"));
    console.log(
      chalk.green(`API server: http://${props.host}:${props.port}/graphql`)
    );
    console.log(
      chalk.green(
        `API documentation: http://${props.host}:${props.port}/graphiql`
      )
    );
  } catch (err) {
    console.error(err);
  }
}
